package de.tfhberlin.panitz.eliza;

/**
  An abstract description of a simple linked list.
**/
public abstract class Li {

    /**
       Constructs a new empty list.
       @return a newly created empty list. 
     **/
  abstract public Li empty();

    /**
       Constructs a new  list by adding an element in front of
       an existing list.
       @param x the first element for the new list.
       @param xs the tail for the new list.
       @return a newly created list. 
     **/
  abstract public Li cons(Object x,Li xs);

    /**
       Tests, if the list contains some element.
       @return true if no element is contained in the list.
     **/
  abstract public boolean isEmpty();

    /**
       Returns the first element of this list.
       @return from element.
     **/
  abstract public Object  head();
    /**
       Returns the list after the first element has been skipped.
       @return the tail (rest) of the list, which follows first element.
     **/
  abstract public Li      tail();


  /** 
     Makes a list of lists of all taillists.
     

     example: for the list ("this","is","a","test")
     this produces:
         (("this","is","a","test"),("is","a","test"),("a","test"),("test"))

     @return a list of lists, which contains all list obtained from
       this list through several calls of tail().
    **/
  abstract public Li tails();

  /** 
    Removes the first i elements from a list.

    @param i a positive number of elements to be skipped
       from the beginning of the list.
    @return this list with i elements droped by tail calls.
   **/
  abstract public Li drop(int i);


  /**
     Reverses the element order of the list.

     example: ("a","b","c") gets ("c","b","a")

     @return a new list with reversed element order of this list.
   **/
  abstract public Li reverse();

  /**
     Produces a list of the words within a String.
    
     The String is splitted at white-space characters. The words
     in the result list will not contain any shite space characters.  

     Example: words("this is a test") produces the list:
           ("this","is","a","test")

     @param str the string, which is to be splitted into words
     @return a list of Strings, which are the words of the input.
   **/
  abstract public Li words(String str);


  /** Makes a list  of String to a String.

    This is a reverse function to words.

    example: unwords for a list: ("this","is","a","test")
             produces the String: "this is a test"

     @return a string which concatenates the Strings of this list with
          whitespace chareacters.
  **/
  abstract public String unwords();

  /** Tests if this list of Strings is a prefix of the other list of Strings.
      
      The lists are supposed to be list of strings. The test is to be made 
      case insensetive.

      example: ("tHIs","is") is prefix of ("this","iS","a","test") 
              if cases are ignored.
 
      @param other the List of Strings, which is tested ,if it extends this
                 List of Strings.
      @return true, if this list prefixes the other list.
  **/
  abstract public boolean isPrefixIgnoreCaseOf(Li/*String*/ other);

    /**
       Mutates this list, by appending first element as last.
     **/
  abstract public void rotate();

    /**
       @return the number of elements in this list.
    **/
  int length(){
    if (isEmpty())return 0;
    return 1+tail().length();
  }

  public String toString(){
    return "("+toStringAux()+")";
  }

  private String toStringAux(){
    if (isEmpty()) return "";
    else if (tail().isEmpty()) return head().toString();
    else return head().toString()+","+tail().toStringAux();
  }

  Li append(Li ys){
    if (isEmpty())return ys;
    return cons(head(),tail().append(ys));
  }

  Li concat(Li ys){ return append(ys);}

  Li filter(FilterCondition cond){
    Li result = empty();

    //test all elements of this list
    for (Li xs=this;!xs.isEmpty();xs=xs.tail()){

      //in case that the condition is true for the element
      if (cond.condition(xs.head())) {
        //then add it to the result
        result = cons(xs.head(),result);
      }
    }
    return result;
  }

  Li quicksort(){
    Li result = empty();
    if (!isEmpty()){ 
      result 
       =   //filter the smaller elements out of the tail
           tail().filter(new LessEqualX((String)head()))
            //sort these
            .quicksort()
            //concatenate it with the sorted 
            //sublist of greater elements
            .append(cons(head()
                   ,tail().filter(new GreaterX((String)head()))
                    .quicksort()
              ));
    }
    return result;
  }

  Li sortBy(Relation rel){
    Li result = empty();
    if (!isEmpty()){ 
      FilterCondition le
       = new OrderingCondition(head(),rel);
      FilterCondition gr
       = new NegativeOrderingCondition(head(),rel);
      result = tail()
                .filter(le)
                .sortBy(rel)
                .append(cons(head()
                              ,tail()
                                 .filter(gr)
                                 .sortBy(rel)));
    }
    return result;
  }
}

